//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#include "Common.h"

// #include "../../../lib/Target/X86/X86.h"
// #include "../../../lib/Target/X86/X86TargetMachine.h"
#include "llvm/Support/Casting.h"
#include "llvm/Support/raw_ostream.h"

#include "llvm/Target/TargetMachine.h"
#include "llvm/CodeGen/MachineRegisterInfo.h"

namespace llvm {
  
  //*** add this class field to enable the physical register set
  const TargetRegisterClass **PhysRegClasses = NULL;
  
  /// Check if the given instruction operator is floating point register kill
  /// operation
  bool isFPREGKill(const TargetMachine& tm, unsigned opcode)
  {
    // if (dynamic_cast<const X86TargetMachine*>(&tm) != NULL) // is the target X86?
    // {
    //   return (opcode == X86::FP_REG_KILL);
    // }
    // else
    // {
    return false;
    // }
  }
  
  void initPhysRegClasses(const TargetRegisterInfo* TRI) {
    if (PhysRegClasses == NULL) {
      unsigned NumRegs = TRI->getNumRegs();
      // add this to enable physical register set
      PhysRegClasses = new const TargetRegisterClass*[NumRegs];

      // Zero initialize.
      for (unsigned I = 0; I != NumRegs; ++I)
        PhysRegClasses[I] = NULL;

      // Populate table.
      for (TargetRegisterInfo::regclass_iterator classit = TRI->regclass_begin();
           classit != TRI->regclass_end();
           ++classit) {
        const TargetRegisterClass *RC = *classit;

        for (unsigned i = 0, e = RC->getNumRegs(); i != e; ++i) {
          unsigned Reg = RC->getRegister(i);
          assert(Reg < NumRegs && "Reg must fit into array");
          PhysRegClasses[Reg] = RC;
        }
      }
    }
  }
  
  /// This is an additional method for the ChaitinBrigges register allocator.
  /// There are physical register that are not part of any register class (for
  /// example, the stack pointer), in this case the function returns NULL.
  const TargetRegisterClass *getRegClass_(const TargetRegisterInfo* tri, unsigned RegNo) {
    assert(TargetRegisterInfo::isPhysicalRegister(RegNo) && "Register number out of range!");
    return PhysRegClasses[RegNo];
  }
  
  const TargetRegisterClass *getRegClass_(MachineFunction& MF, unsigned RegNo) {
    if ( TargetRegisterInfo::isPhysicalRegister(RegNo) ) {
      // tri = MF.getTarget().getRegisterInfo() ... does not have a reg to reg class map method??? 
      return PhysRegClasses[RegNo];
    }
    else {
      const MachineRegisterInfo *mri = &MF.getRegInfo();
      return mri->getRegClass(RegNo);
    }
  }
    
  unsigned compose(const TargetRegisterInfo* TRI, unsigned a, unsigned b) {
    if (!a) return b;
    if (!b) return a;
    return TRI->composeSubRegIndices(a, b);
  }
  
  bool isMoveInstr(const TargetRegisterInfo* TRI, const MachineInstr *MI,
                   unsigned &Src, unsigned &Dst, unsigned &SrcSub, unsigned &DstSub)
  {
    if (MI->isCopy()) {
      Dst = MI->getOperand(0).getReg();
      DstSub = MI->getOperand(0).getSubReg();
      Src = MI->getOperand(1).getReg();
      SrcSub = MI->getOperand(1).getSubReg();
    }
    else if (MI->isSubregToReg()) {
      Dst = MI->getOperand(0).getReg();
      DstSub = compose(TRI, MI->getOperand(0).getSubReg(), MI->getOperand(3).getImm());
      Src = MI->getOperand(2).getReg();
      SrcSub = MI->getOperand(2).getSubReg();
    }
    else {
      return false;
    }
    
    return true;
  }
  
}
