//===--ConstructRegisterClasses.cpp - Constructs register classes for -----===//
//                                    reg alloc
// 
// Anshuman Dasgupta and Jason Eckhardt, MSCP, Rice University
//===----------------------------------------------------------------------===//
//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//

#include "llvm/CodeGen/MachineRegisterInfo.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"

// #include "llvm/ADT/EquivalenceClasses.h"
#include "EquivalenceClasses.h"
#include "Common.h"

#include "ConstructRegisterClasses.h"

namespace llvm {
  ConstructRegisterClasses::ConstructRegisterClasses (const TargetRegisterInfo* RegInfo, 
                                                      MachineFunction& MF)
  {
    // Equivalence classes for sets of registers that can interfere with each other
    // 13 stands for LLVM 1.3
    EquivalenceClasses13<const TargetRegisterClass* > EqRegisterClasses;
    
    TargetRegisterInfo::regclass_iterator classit;
    for (classit = RegInfo->regclass_begin();
         classit != RegInfo->regclass_end();
         ++classit)
    {
      const TargetRegisterClass *RC = *classit;
#if 0
      dbgs() << "\n";
      dbgs() << "Register Class : " << RC->getName() << "\n";
      dbgs() << "\n";
#endif

      // EqRegisterClasses.insert(*classit);
      EqRegisterClasses.addElement(RC);
      
      // All members of regclass form an equivalance class
      for (TargetRegisterClass::iterator rit = RC->begin();
           rit != RC->end(); ++rit) {

        const unsigned Reg = *rit;
#if 0
        {
          unsigned RepReg = getRepresentativeNumber(Reg, RegInfo);
          dbgs() << "Register " << PrintReg(Reg, RegInfo);
          dbgs() << "(" << Reg << ") has Representation Register ";
          dbgs() << PrintReg(RepReg, RegInfo); 
          dbgs() << "(" << RepReg << ").\n";
#if 0
          const unsigned* super = RegInfo->getSuperRegisters(Reg);
          unsigned i = 0;
          if ( super != 0 ) {
            while ( true ) {
              dbgs() << "\tTRI->getSuperRegisters(" << PrintReg(Reg, RegInfo) << ")";
              dbgs() << "[" << int(i) << "] is ";
              dbgs() << PrintReg(super[i], RegInfo); 
              dbgs() << "(" << unsigned(super[i]) << ").\n";
              if ( super[i] == 0 ) break;
              i++;
            }
          }
          else {
            dbgs() << "\tTRI->getSuperRegisters(" << Reg << ") returns NULL.\n";
          }
#endif
        }
#endif

        // Merge the aliases TargetRegisterClasses with its own TRC
        for (const unsigned* aliasIter = RegInfo->getAliasSet(Reg);
             aliasIter && *aliasIter; ++aliasIter)
        {
          unsigned AliasReg = *aliasIter;
          const TargetRegisterClass *AliasRC;

          AliasRC = getRegClass_(RegInfo, AliasReg);
          if ( AliasRC == 0 ) {
            // RIP is in classes { GR64_NOREX, GR64 }
            // RIP is aliased to { EIP, IP }, neither of which is contained in a register class
            continue;
          }
#if 0
          dbgs() << "\tAlias Register " << PrintReg(AliasReg, RegInfo);
          dbgs() << " of class " << AliasRC->getName() << ".\n";;
#endif
          
          EqRegisterClasses.addElement(AliasRC);
          EqRegisterClasses.unionSetsWith(AliasRC, RC);
        }
      }
    }
    
    // Update RegisterClasses
    // EquivalenceClasses<const TargetRegisterClass*>::iterator I,E;
    // for (I = EqRegisterClasses.begin(), E = EqRegisterClasses.end(); I != E; ++I)
    // {
    //   if (!I->isLeader()) continue;
    //   RegisterClasses.push_back(std::set<const TargetRegisterClass*>());
    
    //   EquivalenceClasses<const TargetRegisterClass*>::member_iterator MI, Mend;
    //   MI = EqRegisterClasses.member_begin(I);
    //   Mend = EqRegisterClasses.member_end();
    
    //   if (MI == Mend) continue;
    
    //   //*** Add connection between FR64 and VR128, FR32 and VR128, this is only
    //   //*** special for FR64
    //   if ((*MI)->getID() == 3) {
    //     RegisterClasses.back().insert(RegInfo->getRegClass(25));
    //   } else if ((*MI)->getID() == 2) {
    //     RegisterClasses.back().insert(RegInfo->getRegClass(25));
    //   }
    
    //   for (Mend = EqRegisterClasses.member_end(); MI != Mend; ++MI) {
    //     RegisterClasses.back().insert(*MI);
    //   }
    
    // }
    
    
    std::set<const TargetRegisterClass*>& equivalenceSetReps = EqRegisterClasses.getLeaderSet();
    
    std::set<const TargetRegisterClass*>::iterator repIt;
    for (repIt = equivalenceSetReps.begin();
         repIt != equivalenceSetReps.end();
         ++repIt)
    {
      
      // create a new TargetRegisterClass for each equivalence set
      RegisterClasses.push_back(std::set<const TargetRegisterClass*>());
      RegisterClasses.back().insert(*repIt);
      
      // iterate over the elements of the equivalence class and put them in the
      // register class
      const std::set<const TargetRegisterClass*>& eqElements = EqRegisterClasses.getEqClass(*repIt);
      
      std::set<const TargetRegisterClass* >::iterator elemIt;
      for (elemIt = eqElements.begin(); elemIt != eqElements.end(); ++elemIt) {
        RegisterClasses.back().insert(*elemIt);
      }
    }
  }
}
