//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//


#ifndef GROWABLEDENSEMAP_H_
#define GROWABLEDENSEMAP_H_

namespace llvm {
  
  struct AllRegs2IndexFunctor : std::unary_function<unsigned, unsigned> {
    unsigned operator()(unsigned Reg) const {
      //      return Reg - MRegisterInfo::FirstVirtualRegister;
      return Reg;
    }
  };
  
  struct IdentityFunctor_ : std::unary_function<unsigned, unsigned> {
    unsigned operator()(unsigned Index) const {
      return Index;
    }
  };
  
  template <typename T,  typename ToIndexT = IdentityFunctor_>
  class GrowableDenseMap {
    
    typedef typename ToIndexT::argument_type IndexT;
    
  public:
    typedef std::vector<T> StorageT;
    StorageT storage_;
    T nullVal_;
    ToIndexT toIndex_;
    
    // Must define a null value at construction time
    explicit GrowableDenseMap(const T& val) {
      nullVal_ = val;
    }  
    
    typename StorageT::reference operator[](IndexT n) {
      unsigned index = toIndex_(n);

      // If the value is not found, grow storage
      if (index >= storage_.size()) {
        grow(index);
      }
      
      return storage_[index];
    }
    
    typename StorageT::const_reference operator[](IndexT n) const {
      unsigned index = toIndex_(n);

      // If the value is not found, grow storage
      if (index >= storage_.size()) {
        grow(index);
      }
      
      return storage_[index];
    }
    
    void remove(IndexT n) {
      unsigned index = toIndex_(n);

      assert(index < storage_.size() && "Exceeds bounds");
      storage_[index] = nullVal_;
    }
    
    void clear() {
      storage_.clear();
    }
    
    typename StorageT::size_type size() const {
      return storage_.size();
    }
    
  private:
    void grow(IndexT n) {
      unsigned NewSize = n + 1;
      if (NewSize > storage_.size())
        storage_.resize(NewSize, nullVal_);
    }
    
  };
}
#endif
