//===----------------------------------------------------------------------===//
//This software was produced with support from the Defense Advanced
//Research Projects Agency (DARPA) through AFRL Contract FA8650-09-C-1915. 
//Nothing in this work should be construed as reflecting the official policy or
//position of the Defense Department, the United States government, or
//Rice University.
//
//THIS SOFTWARE HAS BEEN APPROVED FOR PUBLIC RELEASE, UNLIMITED DISTRIBUTION.
//
//===----------------------------------------------------------------------===//


#ifndef __REMATERIALIZATION_H__
#define __REMATERIALIZATION_H__

#include <cstddef>
#include "llvm/ADT/DenseMap.h"
#include "llvm/ADT/DenseSet.h"
#include "llvm/CodeGen/MachineBasicBlock.h"
#include "llvm/Target/TargetRegisterInfo.h"

namespace llvm {
  class MachineFunction;
  class MachineInstr;
  class MachineOperand;
  class TargetInstrInfo;
}

class AvailableDefinitions;

using namespace llvm;

/// Rematerialization support.
class Rematerialization {
  AvailableDefinitions *AvailDef;
  MachineFunction *MF;
  const TargetInstrInfo *TII;
  DenseMap<MachineOperand *, MachineInstr *> RematCandidates;

  // computeAvailDefs - Builds available definitions.
  void computeAvailDefs();

  // findOpportunities - Finds rematerializable instructions.
  void findOpportunities(MachineBasicBlock *BB,
                         DenseSet<unsigned> &ColoredRegs);

  bool isCheaperThanLoadAndSafe(MachineInstr *I);

public:
  Rematerialization() : AvailDef(NULL), MF(NULL) {}

  ~Rematerialization();  // setFunction - Sets the function to operate on.

  void setFunction(MachineFunction *F);


  // findOpportunities - Computes for each instruction which operands are
  // rematerializable.
  // Takes into account that operands of the rematerialized instruction must be
  // in registers.
  void findOpportunities(DenseSet<unsigned> &GuaranteedToGetColor);

  bool isRematerializable(MachineOperand &Oper);

  MachineInstr *getRematerializableInst(MachineOperand &Oper);

  void rematerialize(MachineBasicBlock &BB,
                     MachineBasicBlock::iterator &InsertBefore,
                     unsigned NewReg,
                     MachineOperand &UseToSpill,
                     const TargetRegisterInfo &TRI);

  // clear - Clears any analysis information.
  void clear();
};

#endif
